/*

  Generalized Langton's Ant Viewer 3D

  based on a program of Harald Sanftmann and Daniel Mohr

     by Heiko Hamann 2003

     If you have suggestions, questions, ... please send me an e-mail:
        heiko.hamann@gmx.de
*/



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glut.h>


// size of the world (cube side length)
// be carfull: size^3 bytes are needed
//  100^3: about   1 MB
//  400^3: about  64 MB
//  600^3: about 216 MB
#define size 100


/* 
 * Window properties 
 */
#define WINDOW_WIDTH    1000
#define WINDOW_HEIGHT   800
#define WINDOW_X        0
#define WINDOW_Y        0
#define WINDOW_TITLE    "Generalized Langton's Ant Viewer 3D"

/* 
 * Perspective properties 
 */
#define FOV_ANGLE       60

#define CENTER_X        0.0
#define CENTER_Y        0.0
#define CENTER_Z        0.0

#define VIEWER_X        0.0
#define VIEWER_Y        0.0
#define VIEWER_Z        -2.0

#define UP_X            0.0
#define UP_Y            1.0
#define UP_Z            0.0

#define CLIPPLANE_NEAR  1.0
#define CLIPPLANE_FAR   20.0

 
#define ROTATION_SPEED  5.0

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define PT 1
#define STROKE 2
#define END 3






static GLuint sphereList, cubeList;

typedef struct charpoint {
   GLfloat   x, y;
   int    type;
} CP;

GLfloat rot_x = 0.0, rot_y = 0.0;
GLfloat saved_x, saved_y;


GLfloat mat_ambidiff[] = { 0.7, 0.7, 0.9, 1.0 };
GLfloat mat_ambidiffS[] = { 0.0, 0.0, 0.0, 1.0 };
GLfloat A[] = { 1.0, 0.0, 0.0, 1.0 };
GLfloat B[] = { 0.0, 1.0, 0.0, 1.0 };
GLfloat C[] = { 0.0, 0.0, 1.0, 1.0 };
GLfloat D[] = { 1.0, 1.0, 0.0, 1.0 }; 
GLfloat E[] = { 0.0, 1.0, 1.0, 1.0 };
GLfloat F[] = { 1.0, 0.0, 1.0, 1.0 };
GLfloat G[] = { 1.0, 1.0, 1.0, 1.0 };


GLdouble distance=5.0; //Camera position
GLdouble twist=0.0;
GLdouble elevation=0.0;
GLdouble azimuth=0.0;
GLdouble d_x=0.0;
GLdouble d_y=0.0;
GLdouble d_z=0.0;


int mousex; //Mouse position and button pressed
int mousey;
int mouseButtonNumber;
int axis=0;



// Ant variables
unsigned char universum[size][size][size];
int steps=1;
int dir=0; // direction (0=west,1=north,2=east,3=south,4=up,5=down)
int plane=0;
// plane:
// 0=x-y plane and up is z+,
// 1=x-y plane and up is z-,
// 2=x-z plane and up is y+,
// 3=x-z plane and up is y-,
// 4=y-z plane and up is x+,
// 5=y-z plane and up is x-,


// matrices for changing of direction and plane
int goLeft[6][6]={{3,0,1,2,0,0},
		  {1,2,3,0,0,0},
		  {5,0,4,0,0,2},
		  {4,0,5,0,2,0},
		  {0,4,0,5,3,1},
		  {0,5,0,4,1,3}};
int goRight[6][6]={{1,2,3,0,0,0},
		 {3,0,1,2,0,0},
		 {4,0,5,0,2,0},
		 {5,0,4,0,0,2},
		 {0,5,0,4,1,3},
		 {0,4,0,5,3,1}};
int goUpDir[6]={4,5,3,1,2,0};
int goDownDir[6] ={5,4,1,3,0,2};
int goUpPlane[6][6]={{4,2,5,3,0,0},
		     {4,2,5,3,0,0},
		     {4,0,5,0,1,0},
		     {4,0,5,0,1,0},
		     {0,2,0,3,1,0},
		     {0,2,0,3,1,0}};
int goDownPlane[6][6]={{5,3,4,2,0,0},
		       {5,3,4,2,0,0},
		       {5,0,4,0,0,1},
		       {5,0,4,0,0,1},
		       {0,3,0,2,0,1},
		       {0,3,0,2,0,1}};



int x=size/2,y=size/2,z=size/2;
int stepCount=0;
int inBounds=1;
int rulestring[256];
int rulelength=0;



/*
 * Function prototypes
 */
void usage(void);
void draw_scene(void);
void draw_object(void);
void init(int argc, char **argv, void (*draw)(void));
struct point get_coords(double a, double b);
void vertex(double a, double b);
void set_color(int angle);
void computeAnt(void);
void mousefunc(int button,int state,int x,int y);
void keyboardfunc(unsigned char key,int x,int y);
void mouseMotion(int x, int y);
void mouse(int button, int state, int x, int y);



int main(int argc, char **argv) {

    if(argc==1) {
	printf("\nUsage: You need to give a rule string.\n");
	printf("       Example: ant3d RLUD\n");
	printf("       L,l,0 left\n");
	printf("       R,r,1 right\n");
	printf("       U,u,2 up\n");
	printf("       D,d,3 down\n");
	printf("Type \"ant3d --help\" for more information.\n\n");
	exit(0);
    }
    if(argc>1) {
	if(!strcmp(argv[1],"--help")) {
	    printf("\nGeneralized Langton's Ant Viewer 3D - Help\n\n");
	    printf("Usage: You need to give a rule string.\n");
	    printf("       Examples: \"ant3d LRUD\", \"ant3d lrud\", \"ant3d 0123\",\n");
	    printf("                 \"ant3d Lr2d\", \"ant3d 10\"\n");
	    printf("       0,l,L left\n");
	    printf("       1,r,R right\n");
	    printf("       2,u,U up\n");
	    printf("       3,d,D down\n\n");
	    printf("Use of keys:\n");
	    printf("esc or q quit\n");
	    printf("       d do next step(s)\n");
	    printf("       c clear universum, set ant to center and init direction\n");
	    printf("       a draw axis on/off\n");
	    printf("       1 do 1 step at once by hitting d\n");
	    printf("       2 do 10 steps at once by hitting d\n");
	    printf("       3 do 100 steps at once by hitting d\n");
	    printf("       4 do 1,000 steps at once by hitting d\n");
	    printf("       5 do 10,000 steps at once by hitting d\n");
	    printf("       6 do 100,000 steps at once by hitting d\n");
	    printf("       7 do 1,000,000 steps at once by hitting d\n\n");
	    printf("mouse buttons (hold button and move mouse):\n");
	    printf("left     rotate\n");
	    printf("middle   zoom\n");
	    printf("right    move center\n\n");
	    exit(0);
	}
    }

    /*
     * Init OpenGL and enter the event loop
     */
    init(argc, argv, draw_scene);

    return 0;
}




void computeAnt(void) {
    int count=0;
    int temp=0;

    while((count<steps)&&inBounds) {
	
	count++;
	switch(dir) {
	    case 0: x--;
		if(x<0) inBounds=0;
		break;
	    case 1: y++;
		if(y>=size) inBounds=0;
		break;
	    case 2: x++;
		if(x>=size) inBounds=0;
		break;
	    case 3: y--;
		if(y<0) inBounds=0;
		break;
	    case 4: z++;
		if(z>=size) inBounds=0;
		break;
	    case 5: z--;
		if(z<0) inBounds=0;
		break;
	} // end switch(dir)

	if(inBounds) {
	    switch(rulestring[universum[x][y][z]% rulelength]) {
		case 0: dir=goLeft[plane][dir];
		    break;
		case 1: dir=goRight[plane][dir];
		    break;
		case 2: temp=dir;
		    dir=goUpDir[plane];
		    plane=goUpPlane[plane][temp];
		    break;
		case 3: temp=dir;
		    dir=goDownDir[plane];
		    plane=goDownPlane[plane][temp];
		    break;
	    } // end switch(rulestring)
	    universum[x][y][z]=(universum[x][y][z]+1);
	} // end if inBounds
    } // end while

    if(count==0) 
	printf("Ant is out of Bounds.\n");
    else {
	stepCount+=count;
	printf("%d steps performed.\n", stepCount);
    }
}


void keyboardfunc(unsigned char key,int a,int b) {

    int i,j,k;

    if(key=='q' || key==27) exit(0);
    switch(key) {
	case '1': steps=1; break;
	case '2': steps=10; break;
	case '3': steps=100; break;
	case '4': steps=1000; break;
	case '5': steps=10000; break;
	case '6': steps=100000; break;
	case '7': steps=1000000; break;
	case 'd': computeAnt();
	    glutPostRedisplay();
	    break;
	case 'c': 
	    for(i=0;i<size;i++)
		for(j=0;j<size;j++)
		    for(k=0;k<size;k++)
			universum[i][j][k]=0;
	    dir=0;
	    plane=0;
	    stepCount=0;
	    inBounds=1;
	    x=size/2; y=size/2; z=size/2;
	    glutPostRedisplay();
	    break;
	case 'a': axis=!axis;
	    glutPostRedisplay();
	    break;
    }
}



/*
 * Handle rotations and buffer swapping and call the function draw_object 
 * which does the actual drawing
 */
void draw_scene(void) {
    void polarView(GLdouble distance, GLdouble twist,
		   GLdouble elevation, GLdouble azimuth) {
	glTranslated(0.0, 0.0, -distance);
	glRotated(-twist, 0.0, 0.0, 1.0);
	glRotated(-elevation, 1.0, 0.0, 0.0);
	glRotated(azimuth, 0.0, 1.0, 0.0);
    }
    void printStrokedString(char *s) {
	GLsizei len = strlen(s);
	glCallLists(len, GL_BYTE, (GLbyte *)s);
    }

    void drawAxis3d(void) {
	glColor3f (0.5, 0.5, 0.5); //BEGIN drawAxis3d
	
	glLoadIdentity ();  
	polarView(distance, twist, elevation, azimuth);
	glTranslated(0.0, 0.0, 0.15);
	glColor3f (0.0, 0.0, 1.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambidiffS);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0.0, 0.0, -0.15);
	glVertex3f(0.0, 0.0, 0.0);
	glEnd();
	glutSolidCone(0.02, 0.06, 6, 1); //Z-Axis
	

	glLoadIdentity ();
	polarView(distance, twist, elevation, azimuth);
	glRotated(90.0, 0.0, 1.0, 0.0);
	glTranslated(0.0, 0.0, 0.15);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambidiffS);
	glColor3f (1.0, 0.0, 0.0);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0.0, 0.0, -0.15);
	glVertex3f(0.0, 0.0, 0.0);
	glEnd();
	glutSolidCone(0.02, 0.06, 6, 1); //X-Axis
	
	glLoadIdentity ();
	polarView(distance, twist, elevation, azimuth);
	glRotated(90.0, 0.0, 1.0, 0.0);
	glRotated(90.0, -1.0, 0.0, 0.0);
	glTranslated(0.0, 0.0, 0.15);
	glColor3f (0.0, 0.7, 0.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambidiffS);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0.0, 0.0, -0.15);
	glVertex3f(0.0, 0.0, 0.0);
	glEnd();
	glutSolidCone(0.02, 0.06, 6, 1); //Y-Axis
	
	glLoadIdentity ();  //X-char
	glTranslated(d_x, d_y, d_z);
	polarView(distance, twist, elevation, azimuth);
	glTranslated(0.25, 0.0, 0.0);
	glRotated(-azimuth, 0.0, 1.0, 0.0);
	glRotated(elevation, 1.0, 0.0, 0.0);
	glRotated(twist, 0.0, 0.0, 1.0);
	glTranslated(-d_x, -d_y, -d_z);
	glColor3f (1.0, 0.0, 0.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambidiffS);
	glBegin(GL_LINES);
	glVertex2f(-0.03, -0.03);
	glVertex2f(0.03, 0.03);
	glVertex2f(0.03, -0.03);
	glVertex2f(-0.03, 0.03);
	glEnd();
	glLoadIdentity (); //Y-char
	glTranslated(d_x, d_y, d_z);
	polarView(distance, twist, elevation, azimuth);
	glTranslated(0.0, 0.25, 0.0);
	glRotated(-azimuth, 0.0, 1.0, 0.0);
	glRotated(elevation, 1.0, 0.0, 0.0);
	glRotated(twist, 0.0, 0.0, 1.0);
	glTranslated(-d_x, -d_y, -d_z);
	glColor3f (0.0, 0.7, 0.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambidiffS);
	glBegin(GL_LINES);
	glVertex2f(-0.03, -0.03);
	glVertex2f(0.03, +0.03);
	glVertex2f(0.0, 0.0);
	glVertex2f(-0.03, +0.03);
	glEnd();
	
	glLoadIdentity (); //Z-char
	glTranslated(d_x, d_y, d_z);
	polarView(distance, twist, elevation, azimuth);
	glTranslated(0.0, 0.0, 0.25);
	glRotated(-azimuth, 0.0, 1.0, 0.0);
	glRotated(elevation, 1.0, 0.0, 0.0);
	glRotated(twist, 0.0, 0.0, 1.0);
	glTranslated(-d_x, -d_y, -d_z);
	glColor3f (0.0, 0.0, 1.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambidiffS);
	glBegin(GL_LINE_STRIP);
	glVertex2f(-0.03, 0.03);
	glVertex2f(0.03, 0.03);
	glVertex2f(-0.03, -0.03);
	glVertex2f(0.03, -0.03);
	glEnd();
    }
    
    static GLfloat old_rot_matrix[16];
    static int initialized = 0;
    GLfloat new_rot_matrix[16];

    //BEGIN DISPLAY
    glDisable(GL_LIGHTING); 
    glClearColor (0.9, 0.9, 0.9, 0.0);
    glColor3f (0.0, 0.0, 0.0);
    glLineWidth (0.8);
    
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    glLoadIdentity ();
    glViewport (0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    gluPerspective(48.0, (WINDOW_WIDTH/WINDOW_HEIGHT), 1.5, 20.0);
    glMatrixMode (GL_MODELVIEW);


    polarView(distance, twist, elevation, azimuth);
    glTranslated(d_x, d_y, d_z);
    glEnable(GL_LIGHTING);
    draw_object();
    if(axis)
	drawAxis3d();
    glDisable(GL_LIGHTING);

    glutSwapBuffers();
}

/*
 * Initialize the OpenGL machine
 */
void init(int argc, char **argv, void (*draw)(void)) {

    int i,j,k;
    
    GLfloat light0_pos[] = {100.0, 150.0, 200.0, 1.0};
    GLfloat light0_color[] = {1.0, 1.0, 1.0, 1.0};
    GLfloat ambient_light[] = {0.5, 0.5, 0.5, 1.0};

    
    glutInit(&argc, argv);

    // extract rulestring
    while(argv[1][rulelength]!='\0') {
	switch(argv[1][rulelength]) {
	    case '0': rulestring[rulelength]=0;
		break;
	    case '1': rulestring[rulelength]=1;
		break;
	    case '2': rulestring[rulelength]=2;
		break;
	    case '3': rulestring[rulelength]=3;
		break;
	    case 'l': rulestring[rulelength]=0;
		break;
	    case 'r': rulestring[rulelength]=1;
		break;
	    case 'u': rulestring[rulelength]=2;
		break;
	    case 'd': rulestring[rulelength]=3;
		break;
	    case 'L': rulestring[rulelength]=0;
		break;
	    case 'R': rulestring[rulelength]=1;
		break;
	    case 'U': rulestring[rulelength]=2;
		break;
	    case 'D': rulestring[rulelength]=3;
		break;
	    default: printf("\nUse 0,1,2,3,l,r,u,d,L,R,U,D in rule string only!\n\n");
		exit(0);
	}
	rulelength++;
    }
    
    /* Create a window */
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
    glutInitWindowPosition(WINDOW_X, WINDOW_Y);
    glutCreateWindow(WINDOW_TITLE);

    for(i=0;i<size;i++)
	for(j=0;j<size;j++)
	    for(k=0;k<size;k++)
		universum[i][j][k]=0;

    glClearColor(0.0, 0.0, 0.0, 0.0);

    /* Set up some light sources */
    glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_color);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);

    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);
    
    glEnable(GL_DEPTH_TEST);

    glShadeModel(GL_SMOOTH);

    /* Create a viewing frustum */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(FOV_ANGLE, WINDOW_WIDTH/WINDOW_HEIGHT, CLIPPLANE_NEAR, 
        CLIPPLANE_FAR);
    gluLookAt(VIEWER_X, VIEWER_Y, -VIEWER_Z, CENTER_X, CENTER_Y, CENTER_Z, 
        UP_X, UP_Y, UP_Z);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(VIEWER_X, VIEWER_Y, VIEWER_Z);

    glutDisplayFunc(draw);
    glutMotionFunc (mouseMotion);
    glutMouseFunc (mouse);
    glutKeyboardFunc(keyboardfunc);

    computeAnt();

    glutMainLoop();

    /* Not reached */
}

void mouse(int button, int state, int x, int y) {
  mouseButtonNumber=button;
  mousex=x;
  mousey=y;
}


void mouseMotion(int x, int y) {
  switch (mouseButtonNumber) {
  case GLUT_LEFT_BUTTON:
    elevation=elevation-(y-mousey)/3.0;
    if (elevation>180) elevation=elevation-360.0;
    if (elevation<-180) elevation=elevation+360.0;
    if ((elevation<90.0) & (elevation>-90.0)) {
      azimuth=azimuth-(x-mousex)/3.0;
      if (azimuth>180) azimuth=azimuth-360.0;
      if (azimuth<-180) azimuth=azimuth+360.0;
    }
    break;
  case GLUT_MIDDLE_BUTTON:
    distance=distance+(y-mousey)/30.0;
    break;
  case GLUT_RIGHT_BUTTON:
    d_x=d_x+
      sin(elevation*3.14/180)*sin(azimuth*3.14/180)*(y-mousey)/100.0+
      cos(azimuth*3.14/180)*(x-mousex)/100.0;
    d_y=d_y-cos(elevation*3.14/180)*(y-mousey)/100.0;
    d_z=d_z+sin(azimuth*3.14/180)*(x-mousex)/100.0-
      sin(elevation*3.14/180)*cos(azimuth*3.14/180)*(y-mousey)/100.0;
    break;
  }
  mousex=x;
  mousey=y;
  glutPostRedisplay();
}


void draw_object(void) {

    float d=1.0/((float)size);
    int i,j,k;
    float x1, x2, x3;

    glBegin(GL_QUADS);

    for(i=1;i<size-1;i++) {
	x1=(2*((float)i-((float)size/2.0)))/(float)size;
	for(j=1;j<size-1;j++) {
	    x2=(2*((float)j-((float)size/2.0)))/(float)size;
	    for(k=1;k<size-1;k++) {

		x3=(2*((float)k-((float)size/2.0)))/(float)size;

		switch(universum[i][j][k]%rulelength) {
		    case 0: break;
		    case 1: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, A);
			break;
		    case 2: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, B);
			break;
		    case 3: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, C);
			break;
		    case 4: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, D);
			break;
		    case 5: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, E);
			break;
		    case 6: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, F);
			break;
		    default: glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, G);
		}


		if((universum[i][j][k]%rulelength)!=0) {
		    
		    if(!universum[i][j][k+1]) {
			glNormal3f( 0.0f, 0.0f, 1.0f);
			glVertex3f(-d+x1, -d+x2,  d+x3); // v1
			glVertex3f( d+x1, -d+x2,  d+x3); // v2
			glVertex3f( d+x1,  d+x2,  d+x3); // v3
			glVertex3f(-d+x1,  d+x2,  d+x3); // v4
		    }
		    if(!universum[i+1][j][k]) {
			glNormal3f( 1.0f, 0.0f, 0.0f);
			glVertex3f( d+x1, -d+x2,  d+x3); // v2
			glVertex3f( d+x1, -d+x2, -d+x3); // v6
			glVertex3f( d+x1,  d+x2, -d+x3); // v5
			glVertex3f( d+x1,  d+x2,  d+x3); // v3
		    }
		    if(!universum[i][j+1][k]) {
			glNormal3f( 0.0f, 1.0f, 0.0f);
			glVertex3f( d+x1,  d+x2,  d+x3); // v3    
			glVertex3f( d+x1,  d+x2, -d+x3); // v5   
			glVertex3f(-d+x1,  d+x2, -d+x3); // v7
			glVertex3f(-d+x1,  d+x2,  d+x3); // v4
		    }
		    if(!universum[i][j-1][k]) {
			glNormal3f( 0.0f,-1.0f, 0.0f);
			glVertex3f( d+x1, -d+x2,  d+x3); // v2
			glVertex3f( d+x1, -d+x2, -d+x3); // v6
			glVertex3f(-d+x1, -d+x2, -d+x3); // v8
			glVertex3f(-d+x1, -d+x2,  d+x3); // v1
		    }
		    if(!universum[i][j][k-1]) {
			glNormal3f( 0.0f, 0.0f,-1.0f);
			glVertex3f(-d+x1, -d+x2, -d+x3); // v8
			glVertex3f( d+x1, -d+x2, -d+x3); // v6
			glVertex3f( d+x1,  d+x2, -d+x3); // v5   
			glVertex3f(-d+x1,  d+x2, -d+x3); // v7
		    }
		    if(!universum[i-1][j][k]) {
			glNormal3f(-1.0f, 0.0f, 0.0f); 
			glVertex3f(-d+x1, -d+x2,  d+x3); // v1
			glVertex3f(-d+x1, -d+x2, -d+x3); // v8
			glVertex3f(-d+x1,  d+x2, -d+x3); // v7
			glVertex3f(-d+x1,  d+x2,  d+x3); // v4
		    }
		}
	    }
	}
    }
    glEnd();
}























































